define(["require", "exports", "loggerhythm", "uuid", "@essential-projects/http", "@process-engine/consumer_api_contracts", "./accessors/external_accessor", "./external_task_api_client_service"], function (require, exports, loggerhythm_1, uuid, http_1, consumer_api_contracts_1, external_accessor_1, external_task_api_client_service_1) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    const logger = loggerhythm_1.Logger.createLogger('processengine:external_task:worker');
    class ExternalTaskWorker {
        constructor(processEngineUrl, identity, topic, maxTasks, longpollingTimeout, processingFunction) {
            this._workerId = uuid.v4();
            this.lockDuration = 30000;
            this._pollingActive = false;
            this.processEngineUrl = processEngineUrl;
            this.identity = identity;
            this.topic = topic;
            this.maxTasks = maxTasks;
            this.longpollingTimeout = longpollingTimeout;
            this.processingFunction = processingFunction;
            this.initialize();
        }
        get workerId() {
            return this._workerId;
        }
        get pollingIsActive() {
            return this._pollingActive;
        }
        start() {
            this._pollingActive = true;
            this.processExternalTasks();
        }
        stop() {
            this._pollingActive = false;
        }
        initialize() {
            const httpClient = new http_1.HttpClient();
            httpClient.config = { url: this.processEngineUrl };
            const externalAccessor = new external_accessor_1.ExternalTaskApiExternalAccessor(httpClient);
            this.externalTaskClient = new external_task_api_client_service_1.ExternalTaskApiClientService(externalAccessor);
        }
        async processExternalTasks() {
            while (this.pollingIsActive) {
                const externalTasks = await this.fetchAndLockExternalTasks(this.identity, this.topic, this.maxTasks, this.longpollingTimeout);
                if (externalTasks.length === 0) {
                    await this.sleep(1000);
                    continue;
                }
                const executeTaskPromises = [];
                for (const externalTask of externalTasks) {
                    executeTaskPromises.push(this.executeExternalTask(this.identity, externalTask));
                }
                await Promise.all(executeTaskPromises);
            }
        }
        async fetchAndLockExternalTasks(identity, topic, maxTasks, longpollingTimeout) {
            try {
                return await this
                    .externalTaskClient
                    .fetchAndLockExternalTasks(identity, this.workerId, topic, maxTasks, longpollingTimeout, this.lockDuration);
            }
            catch (error) {
                logger.error('An error occured during fetchAndLock!', error.message, error.stack);
                // Returning an empty Array here, since "waitForAndHandle" already implements a timeout, in case no tasks are available for processing.
                // No need to do that twice.
                return [];
            }
        }
        async executeExternalTask(identity, externalTask) {
            try {
                const lockExtensionBuffer = 5000;
                const interval = setInterval(async () => this.extendLocks(identity, externalTask), this.lockDuration - lockExtensionBuffer);
                const result = await this.processingFunction(externalTask);
                clearInterval(interval);
                if (result instanceof consumer_api_contracts_1.DataModels.ExternalTask.ExternalTaskBpmnError) {
                    await this.externalTaskClient.handleBpmnError(identity, this.workerId, externalTask.id, result.errorCode);
                }
                else if (result instanceof consumer_api_contracts_1.DataModels.ExternalTask.ExternalTaskServiceError) {
                    await this.externalTaskClient.handleServiceError(identity, this.workerId, externalTask.id, result.errorMessage, result.errorDetails);
                }
                else {
                    const resultAsSuccess = result;
                    await this
                        .externalTaskClient
                        .finishExternalTask(identity, this.workerId, externalTask.id, resultAsSuccess.result);
                }
            }
            catch (error) {
                logger.error('Failed to execute ExternalTask!', error.message, error.stack);
                await this.externalTaskClient.handleServiceError(identity, this.workerId, externalTask.id, error.message, '');
            }
        }
        async extendLocks(identity, externalTask) {
            try {
                await this.externalTaskClient.extendLock(identity, this.workerId, externalTask.id, this.lockDuration);
            }
            catch (error) {
                // This can happen, if the lock-extension was performed after the task was already finished.
                // Since this isn't really an error, a warning suffices here.
                logger.warn(`An error occured while trying to extend the lock for ExternalTask ${externalTask.id}`, error.message, error.stack);
            }
        }
        async sleep(milliseconds) {
            return new Promise((resolve) => {
                setTimeout(() => { resolve(); }, milliseconds);
            });
        }
    }
    exports.ExternalTaskWorker = ExternalTaskWorker;
});
//# sourceMappingURL=external_task_worker.js.map